package org.openxma.addons.ui.table.customizer.xma.client;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.openxma.addons.ui.table.customizer.common.ColumnMetaData;
import org.openxma.addons.ui.table.customizer.common.CommonColumnModel;
import org.openxma.addons.ui.table.customizer.common.CommonDataModel.CodDatatype;
import org.openxma.addons.ui.table.customizer.common.CommonFilterModel;
import org.openxma.addons.ui.table.customizer.common.CommonTableModel;

import at.spardat.enterprise.fmt.ABcdFmtDefault;
import at.spardat.enterprise.fmt.ABcdFmtPattern;
import at.spardat.enterprise.fmt.ABcdFmtRange;
import at.spardat.enterprise.fmt.ADateFmtMediumSmart;
import at.spardat.enterprise.fmt.ATimeStampFmt;
import at.spardat.enterprise.fmt.IFmt;
import at.spardat.xma.mdl.paging.PagingWMClient;
import at.spardat.xma.mdl.table.ITableWMClient;
import at.spardat.xma.mdl.table.TableExternalSorterClient;
import at.spardat.xma.mdl.table.TableLayoutManager;
import at.spardat.xma.mdl.table.TableUIDelegateClient;
import at.spardat.xma.mdl.table.XMATableColumn;
import at.spardat.xma.rpc.RemoteCallClient;
import at.spardat.xma.rpc.RemoteReply;
import at.spardat.xma.session.XMASessionClient;

/**
 * This class implements the client part of TableCustomizerComp.
 * 
 */
public class TableCustomizerComp extends TableCustomizerCompGen implements ITableCustomizer, TableExternalSorterClient {
    private TableLayoutManager tlm;
    private ITableWMClient tableWMClient;
    private String tableName;
    private Map<String, ColumnMetaData> columnMap = new HashMap<String, ColumnMetaData>();
    private PagingWMClient pagingWMClient;
    private String appCode;
    private CommonTableModel commonTableModel;
    public final static short IMG_SORT_ASC = 100;
    public final static short IMG_SORT_DESC = 101;
    public final static short IMG_FILTER = 102;
    public final static short IMG_FILTER_DELETE = 103;
    public final static short IMG_FILTER_SORT_ASC = 104;
    public final static short IMG_FILTER_SORT_DESC = 105;
    
    /**
     * Initializes the TableCustomizerComp. This method is called by the XMA-Runtime.
     * 
     * @param session
     *            the XMASession, this TableCustomizerComp belongs to.
     */
    public TableCustomizerComp(XMASessionClient session) {
        super(session);
        setMainPage(new TableCustomizerPage(this));
    }

    /**
     * Invokes the TableCustomizerComp. This method reads the key-value-pairs of the input, shows the windows of the
     * TableCustomizerComp and returns the output as key-value-pairs. This method is called by the XMA-Runtime.
     * 
     * @param parentComposite
     *            a composite of the calling page. In most cases it will be a shell. It my be null.
     */
    public void invoke(Composite parentComposite) {
        super.invoke(parentComposite);
    }

    public boolean applySettings() {
        return applySettings(false);
    }

    public boolean applySettings(boolean applyOnServer) {
        if (commonTableModel == null) {
            commonTableModel = commonModelFromTable(tableName);
        }
        applyTableModelOnUI(commonTableModel);

        if (applyOnServer) {
            pushTableModel(commonTableModel);
            pagingWMClient.reload();
        }
        return true;
    }

    /**
     * this method applies the visibility/order settings on the GUI. if the user resized any column widths on the table,
     * the table model is updated regarding to the new column sizes. (MANTIS 45565)
     * 
     * @param tableModel
     * @author s4162
     */
    protected void applyTableModelOnUI(CommonTableModel tableModel) {
        commonTableModel = tableModel;
        for (int uiIndex = 0; uiIndex < getTable().getColumnCount(); uiIndex++) {
            TableColumn uiColumn = getTable().getColumn(uiIndex);
            CommonColumnModel cm = tableModel.getColumnModelByColIndex(getTableWMClient().uiColumnIndexToModelColumnIndex(uiIndex));
            if (cm == null) {
                if (pagingWMClient.getSortingColumn() != uiIndex) {
                    uiColumn.setImage(null);
                } else if (pagingWMClient.getAscending()) {
                    uiColumn.setImage(getImage(TableCustomizerComp.IMG_SORT_ASC));
                } else {
                    uiColumn.setImage(getImage(TableCustomizerComp.IMG_SORT_DESC));
                }
            } else {
                if (cm.getFilterModel() != null && cm.getFilterModel().size() > 0) {
                    if (cm.getSortOrder() != 0 || pagingWMClient.getSortingColumn() != -1) { // only show filter icon with sort arrow, if it is the primary sort order
                        uiColumn.setImage(getImage(TableCustomizerComp.IMG_FILTER));
                    } else if (cm.isAscending()) {
                        uiColumn.setImage(getImage(TableCustomizerComp.IMG_FILTER_SORT_ASC));
                    } else {
                        uiColumn.setImage(getImage(TableCustomizerComp.IMG_FILTER_SORT_DESC));
                    }
                } else {
                    if (cm.getSortOrder() != 0 || pagingWMClient.getSortingColumn() != -1) {
                        getTable().getColumn(uiIndex).setImage(null);
                    } else if (cm.isAscending()) {
                        uiColumn.setImage(getImage(TableCustomizerComp.IMG_SORT_ASC));
                    } else {
                        uiColumn.setImage(getImage(TableCustomizerComp.IMG_SORT_DESC));
                    }
                }
                if (cm.getDefWith() == null || cm.getDefWith() == -1) {
                    getTableLayoutManager().setMinWidth(uiIndex, cm.getDefMinWith());
                    getTableLayoutManager().setPercent(uiIndex, cm.getDefPercent());
                } else {
                    getTableLayoutManager().setAbsolutWidth(uiIndex, cm.getDefWith());
                }
                if (cm.isVisible()) {
                    getTableWMClient().getColumn(cm.getNumColumnIndex()).setVisible(cm.isVisible());
                    getTableLayoutManager().layout();
                } else {
                    getTableLayoutManager().collapseColumn(uiIndex);
                }
            }
        }
        
        // int[] columnOrder = tableModel.getColumnOrder();
        int[] currentSwtOrder = getTable().getColumnOrder();
        int[] finalOrder = new int[currentSwtOrder.length];

        List<CommonColumnModel> cms = tableModel.getColumnModels();
        Map<Integer, Boolean> metaMap = new HashMap<Integer, Boolean>();
        for (int i = 0; i < cms.size(); i++) {
            CommonColumnModel columnModelByColIndex = cms.get(i);
            int swtColumnOrder = getTableWMClient()
                    .modelColumnIndexToUIColumnIndex(columnModelByColIndex.getNumOrder());
            finalOrder[swtColumnOrder] = getTableWMClient().modelColumnIndexToUIColumnIndex(
                    columnModelByColIndex.getNumColumnIndex());
            metaMap.put(swtColumnOrder, Boolean.TRUE);
        }
        for (int i = 0; i < finalOrder.length; i++) {
            if (metaMap.get(i) == null) {
                finalOrder[i] = currentSwtOrder[i];
            }
        }

        getTable().setColumnOrder(finalOrder);
    }

    public void setTableLayoutManager(TableLayoutManager tlm) {
        this.tlm = tlm;
    }

    public void setTableWMClient(ITableWMClient tableWMClient) {
        this.tableWMClient = tableWMClient;
    }

    public void customize() {
        try {
            invoke(getComposite());
        } finally {
            // dispose();
        }
    }

    public FilterPage<?> callFilterDialog(TableColumn column) {
        TableCustomizerPage customizerPage;
        Composite composite = getComposite();
        if (composite instanceof Shell) {
            customizerPage = new TableCustomizerPage(this, (Shell) composite);
        } else {
            customizerPage = new TableCustomizerPage(this);
        }
        FilterPage callFilterDialog = customizerPage.callFilterDialog(getColumnAttributeByLabel(column.getText()));
        if (callFilterDialog.getExitstatus()) {
            customizerPage.applyOnSession();
        }
        return callFilterDialog;
    }

    public TableLayoutManager getTableLayoutManager() {
        return tlm;
    }

    public Table getTable() {
        return (Table) (getTableUIDeligate()).getUIControl();
    }

    public ITableWMClient getTableWMClient() {
        return tableWMClient;
    }

    protected at.spardat.xma.mdl.table.TableUIDelegateClient getTableUIDeligate() {
        return (TableUIDelegateClient) getTableWMClient().getUIDelegate();
    }

    public String getFQTableName() {
        return tableName;
    }

    public void setFQTableName(String tableName) {
        this.tableName = tableName;
        setFQTableNameProperty(tableName);
    }

    public void setColumnMap(Map<String, ColumnMetaData> columnMap) {
        this.columnMap = columnMap;
    }

    public Map<String, ColumnMetaData> getColumnMap() {
        return columnMap;
    }

    public void register() {
        newRemoteCall("register").execute();

    }

    public void init() {
        registerImage(TableCustomizerComp.IMG_SORT_ASC,
                "org/openxma/addons/ui/table/customizer/xma/client/images/sort_asc.png");
        registerImage(TableCustomizerComp.IMG_SORT_DESC,
                "org/openxma/addons/ui/table/customizer/xma/client/images/sort_desc.png");
        registerImage(TableCustomizerComp.IMG_FILTER,
                "org/openxma/addons/ui/table/customizer/xma/client/images/filter.png");
        registerImage(TableCustomizerComp.IMG_FILTER_DELETE,
                "org/openxma/addons/ui/table/customizer/xma/client/images/filter_delete.png");
        registerImage(TableCustomizerComp.IMG_FILTER_SORT_ASC,
                "org/openxma/addons/ui/table/customizer/xma/client/images/filter_asc.png");
        registerImage(TableCustomizerComp.IMG_FILTER_SORT_DESC,
                "org/openxma/addons/ui/table/customizer/xma/client/images/filter_desc.png");

        RemoteCallClient newRemoteCall = newRemoteCall("init");
        newRemoteCall.setParameter(0, getColumnMap());
        newRemoteCall.setParameter(1, commonModelFromTable(getFQTableName()));
        RemoteReply reply = newRemoteCall.execute();
        commonTableModel = (CommonTableModel) reply.getParameter(0);
        tableWMClient.setSortIndicator(false);
        tableWMClient.setExternalSorter(this);
        
    }
    
    public String getColumnAttributeByLabel(String columnLabel) {
        for (Entry<String, ColumnMetaData> entry : getColumnMap().entrySet()) {
            String displayText = entry.getValue().getDisplayText();
            if (columnLabel.equals(displayText)) {
                return entry.getKey();
            }
        }
        throw new IllegalArgumentException("the given label value is not defined in the column mappings: "
                + columnLabel);
    }

    protected void updateTableModelColumnSizesFromUI(CommonTableModel commonTableModel) {
        this.commonTableModel = commonTableModel;
        if (commonTableModel != null) {
            for (int i = 0; i < commonTableModel.getColumnModels().size(); i++) {

                CommonColumnModel cm = commonTableModel.getColumnModels().get(i);
                Integer numColumnIndex = cm.getNumColumnIndex();
                XMATableColumn column = getTableWMClient().getColumn(numColumnIndex);
                int uiIndex = getTableWMClient().modelColumnIndexToUIColumnIndex(numColumnIndex);

                if (cm.getDefWith() == null || cm.getDefWith() == -1) {
                    cm.setDefPercent(getTableLayoutManager().getPercent(uiIndex));
                    cm.setDefMinWith(getTableLayoutManager().getMinWidth(uiIndex));
                } else {
                    if (cm.isVisible()) {
                        int currentWidth = getTable().getColumn(uiIndex).getWidth();
                        if (currentWidth == 0) {
                            // an invisible column becomes visible
                            // the model must not be updated
                        } else {
                            // the size of an visible column my have changed
                            // so the model must be updated
                            cm.setDefWith(currentWidth);
                        }
                    }
                    // cm.setDefWith(getTableLayoutManager().getAbsolutWidth(uiIndex));
                }
                if (cm.isVisible()) {
                    column.setVisible(cm.isVisible());
                } else {
                    getTableLayoutManager().collapseColumn(uiIndex);
                }
            }
        }
    }

    protected CommonTableModel commonModelFromTable(String fqTableName) {
        CommonTableModel ctm = new CommonTableModel();
        ctm.setNamTable(fqTableName);
        ctm.setAppCode(getAppCode());
        // load all column labels into model
        Map<String, ColumnMetaData> columnMap = getColumnMap();
        Iterator<String> colIterator = columnMap.keySet().iterator();
        while (colIterator.hasNext()) {
            String attrKey = colIterator.next();
            String attrLabel = (String) columnMap.get(attrKey).getDisplayText();
            String attrTypename = (String) columnMap.get(attrKey).getTypeName();
            for (int i = 0; i < getTableWMClient().getColumnCount(); i++) {
                int swtIndex = getTableWMClient().modelColumnIndexToUIColumnIndex(i);
                if (swtIndex == -1) {
                    // not a column in the widget, do not handle this column in
                    // the dialog
                } else {
                    TableColumn column = getTable().getColumn(swtIndex);
                    int colIndexInTableModel = i;
                    if (column.getText().equals(attrLabel)) {
                        // column at wiget index i found
                        String columnAttributeByLabel = getColumnAttributeByLabel(column.getText());
                        CommonColumnModel cmdl = new CommonColumnModel();
                        cmdl.setCodDataType(CodDatatype.getCodDatatypeByNativeName(attrTypename));
                        cmdl.setNamColumn(columnAttributeByLabel);
                        cmdl.setDefPercent(getTableLayoutManager().getPercent(swtIndex));
                        cmdl.setDefWith(getTableLayoutManager().getAbsolutWidth(swtIndex));
                        cmdl.setNumColumnIndex(i);
                        // cmdl.setNumOrder(i);
                        cmdl.setNumOrder(getTableWMClient().uiColumnIndexToModelColumnIndex(
                                getTable().getColumnOrder()[swtIndex]));
                        cmdl.setTxtColumn(attrLabel);
                        cmdl.setVisible(getTableWMClient().getColumn(colIndexInTableModel).isVisible());
                        cmdl.setResizeable(getTable().getColumn(swtIndex).getResizable());
                        cmdl.setDefMinWith(getTableLayoutManager().getMinWidth(swtIndex));
                        ctm.addColumnModel(cmdl);
                        IFmt formatter = getTableWMClient().getColumn(colIndexInTableModel).getFormatter();
                        String fmtStr = null;
                        if (formatter != null) {
                            if (formatter instanceof ABcdFmtDefault) {
                                int maxBeforeC = ((ABcdFmtDefault) formatter).getMaxBeforeC();
                                int maxAfterC = ((ABcdFmtDefault) formatter).getMaxAfterC();
                                fmtStr = maxBeforeC + "," + maxAfterC;
                            } else if (formatter instanceof ABcdFmtRange) {
                                int maxBeforeC = ((ABcdFmtRange) formatter).getMaxBeforeC();
                                int maxAfterC = ((ABcdFmtRange) formatter).getMaxAfterC();
                                fmtStr = maxBeforeC + "," + maxAfterC;
                            } else if (formatter instanceof ABcdFmtPattern) {
                                int maxBeforeC = ((ABcdFmtPattern) formatter).getMaxBeforeC();
                                int maxAfterC = ((ABcdFmtPattern) formatter).getMaxAfterC();
                                fmtStr = maxBeforeC + "," + maxAfterC;
                            } else if (formatter instanceof ADateFmtMediumSmart) {
                                fmtStr = ((ADateFmtMediumSmart) formatter).getPattern();
                            } else if (formatter instanceof ATimeStampFmt) {
                                //fmtStr = ((ATimeStampFmt) formatter).getPattern(); //NOT IMPLEMENTED!
                            }
                            // TODO: custom formatters are not considered=>
                            // handle this use case
                            cmdl.setTxtFormatterPattern(fmtStr);
                        }
                        break;
                    }
                }

            }
        }
        return ctm;
    }

    public void setPagingWMClient(PagingWMClient pagingWMClient) {
        this.pagingWMClient = pagingWMClient;

    }

    public PagingWMClient getPagingWMClient() {
        return this.pagingWMClient;
    }

    public String getAppCode() {
        return appCode;
    }

    public void setAppCode(String appCode) {
        this.appCode = appCode;
    }

    @Deprecated
    /**
     * @deprecated As of version 5.0.4 applySettings(true) can be used to push a changed tablemodel (getTableModel()) to the server and reloading the data.
     * @see TableCustomizerComp#setFilter(String, int, boolean, List)
     */
    public void setFilter(String fqTableName, int columnId, boolean isAnd, List<CommonFilterModel> filters) {
        CommonTableModel commonModelFromTable = commonModelFromTable(fqTableName);
        RemoteCallClient newRemoteCall = newRemoteCall("setFilter");
        newRemoteCall.setParameter(0, commonModelFromTable);
        newRemoteCall.setParameter(1, Integer.valueOf(columnId));
        newRemoteCall.setParameter(2, Boolean.valueOf(isAnd));
        newRemoteCall.setParameter(3, filters);
        newRemoteCall.execute();
        getPagingWMClient().reload();
    }

    public void removeAllFilter() {
        newRemoteCall("removeAllFilter").execute();
    }

    public CommonTableModel getTableModel() {
        return commonTableModel;
    }

    /**
     * pushes the current settings to the server without persisting them
     * 
     */
    public void pushTableModel(CommonTableModel tableModel) {
        RemoteCallClient push = newRemoteCall("onPushTableModel");
        push.setParameter(0, tableModel);
        push.execute();
    }

    /**
     * Interface sort implementation redirects to version where ascending is nullable, so a sortorder can be removed.
     */
    public void sort(int sortingColumn, boolean ascending) {
        sort(sortingColumn, Boolean.valueOf(ascending));
    }
    //TODO: Weird mix of pagingWMClient and tableWMClient sorting features - maybe try and remove all references to pagingWMClient for sorting - would also need to adapt that on other pagingWMClient referencing methods
    public void sort(int sortingColumn, Boolean ascending) {
        CommonColumnModel columnModel = commonTableModel.getColumnModelByColIndex(getTableWMClient().uiColumnIndexToModelColumnIndex(sortingColumn));
        if (columnModel != null) {
            pagingWMClient.setSortingColumn((short) -1);
            if (ascending == null) {
                rearrangeSortOrders(columnModel, Integer.MAX_VALUE);
                columnModel.setAscending(null);
                tableWMClient.sortNatural();
            } else {
                rearrangeSortOrders(columnModel, 0);
                //columnModel.setAscending(!columnModel.isAscending());
                columnModel.setAscending(ascending);
            }
            pushTableModel(commonTableModel);
            pagingWMClient.start();
        } else { // if column not managed by customizer, redirect to pagingWMClient
            if (ascending == null) {                
                pagingWMClient.setSortingColumn((short) -1);
                tableWMClient.sortNatural();
            } else {
                tableWMClient.setExternalSorter(null);
                tableWMClient.sort(sortingColumn);
                if (tableWMClient.isSortingColumnAscending() != ascending) { //ugly but no other solution, since you cant set ascending
                    tableWMClient.sort(sortingColumn);
                }
                pagingWMClient.setSortingColumn((short) sortingColumn);
                pagingWMClient.setAscending(ascending);
                tableWMClient.setExternalSorter(this);
            }
        }
        applyTableModelOnUI(commonTableModel);
    }

    private void rearrangeSortOrders(CommonColumnModel columnModel, int newSortOrder) {
        List<CommonColumnModel> sortableColumnsSortedByNumSort = commonTableModel
                .getSortableColumnsSortedByNumSort();
        for (CommonColumnModel sortableColumnModel : sortableColumnsSortedByNumSort) {
            if (newSortOrder != Integer.MAX_VALUE) {
                if (sortableColumnModel == columnModel) {
                    break;
                }
                sortableColumnModel.setSortOrder(sortableColumnModel.getSortOrder() + 1);
            } else {
                if (sortableColumnModel != columnModel && sortableColumnModel.getSortOrder() > columnModel.getSortOrder()) {
                    sortableColumnModel.setSortOrder(sortableColumnModel.getSortOrder() - 1);
                }
            }
        }
        columnModel.setSortOrder(newSortOrder);
    }

}